/* To reproduce bug:
 *
 * Run ./zmb then type ctrl-C anytime after the barrier message, but before
 * the ending message.  This is likely to leave a page in the locked state,
 * even though the process may terminate.
 *
 * Next try unmounting the filesystem.  If any pages are in that state,
 * umount will hang and a backtrace from crash will show the process
 * has called schedule() from ___wait_on_page().
 * That's the bug!
 */
#include <pthread.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>

#define NUM_THREADS 2

/* Create 1G zmbtestfile0 and zmbtestfile1 with
 * dd if=/dev/zero of=zmbtestfile0 bs=1048576 count=1024
 * (They could be smaller if necessary).
 */
#define TESTFILE "/p/ba18/garlick/zmbtestfile%d"

static struct {
    pthread_attr_t attr;
    pthread_t thd;
} thd[NUM_THREADS];

static pthread_mutex_t start_lock = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t start_cond = PTHREAD_COND_INITIALIZER;
static int start_flag = 0;

static void *_thread(void *arg)
{
    unsigned char buf[8192];
    int my_threadnum = (int) arg;
    int fd, res;
    size_t size;
    char filename[1024];

    printf("%d: starting\n", my_threadnum);

    /* barrier */
    pthread_mutex_lock(&start_lock);
    while (start_flag == 0)
        pthread_cond_wait(&start_cond, &start_lock);
    pthread_mutex_unlock(&start_lock);
    printf("%d: barrier\n", my_threadnum);

    sprintf(filename, TESTFILE, my_threadnum);
    if ((fd = open(filename, O_RDONLY)) < 0) {
        perror(filename);
        goto done;
    }
    size = 0;
    while ((res = read(fd, buf, sizeof(buf))) > 0)
        size += res;
        ;
    close(fd);
    if (res < 0) {
        perror(filename);
        goto done;
    }

done:
    printf("%d: ending (read %ld Mbytes)\n", my_threadnum, size/(1024L*1024));
    return NULL;
}


int
main(int argc, char *argv[])
{
    int i;

    for (i = 0; i < NUM_THREADS; i++) {
        pthread_attr_init(&thd[i].attr);
        /*pthread_attr_setdetachstate(&thd[i].attr, PTHREAD_CREATE_DETACHED);*/
        pthread_create(&thd[i].thd, &thd[i].attr, _thread, (void *) i);
    }

    /* signal for barrier */
    pthread_mutex_lock(&start_lock);
    start_flag = 1;
    pthread_cond_broadcast(&start_cond);
    pthread_mutex_unlock(&start_lock);

    for (i = 0; i < NUM_THREADS; i++) {
        pthread_join(thd[i].thd, NULL);
    }

    printf("main exiting \n");
    exit(0);
}

/*
 * vi:tabstop=4 shiftwidth=4 expandtab
 */
